from multipledispatch import dispatch
import json

class Utils:
    def __init__(self, attributes) -> None:
        self.attributes = attributes

    def deserialize(self, _json):
        data = json.loads(_json)
        [self.validate_value(self, data, attrib) for attrib in self.attributes]
        return self

    def serialize(self):
        data = {}
        object = self.__dict__
        [self.validate_value(data, object, attrib) for attrib in self.attributes]
        return json.dumps(data)
    
    @dispatch(dict, dict, str)
    def validate_value(self, target_object, object, attribute):
        if object.get(attribute):
            target_object[attribute] = object[attribute]
            return target_object
    
    @dispatch(object, dict, str)
    def validate_value(self, target_object, object, attribute):
        if object.get(attribute):
            setattr(target_object, attribute, object[attribute])
            return target_object
            
    def validate_type(self, value, types):
        return value in types

class BaseModel(Utils):
    def __init__(self, attributes, space_id=None, kibana=None, _json=None) -> None:
        super().__init__(attributes)
        self.space_id = "" if not space_id else f's/{space_id}'
        self.kibana = kibana
        self._json = _json

    def requester(self, **kwargs):
        return self.kibana.requester(**kwargs)

    def url(self, *args):
        return self.kibana.url(self.kibana.base_url, self.space_id, *args)


import json
import secrets
import os
from .base import BaseModel, Utils

class Space(BaseModel):
    def __init__(self, id=None, name=None, description=None, color=None, initials=None, disabledFeatures=None, _reserved=None, kibana=None) -> None:
        super().__init__(["id", "name","description", "color", "initials", "disabledFeatures", "_reserved"], kibana=kibana)
        self.create_url = "api/spaces/space"
        self.id = id
        self.name = name
        self.description = description
        self.color = color
        self.initials = initials
        self.disabledFeatures = disabledFeatures
        self._reserved = _reserved

    def create(self):
        url = self.url(self.create_url)
        data = self.serialize()
        response = self.requester(url=url, method="post", data=data)
        return response

class Object(BaseModel):
    def __init__(self, space_id=None, kibana=None, attribs={}, type="", references={}) -> None:
        super().__init__([], space_id=space_id, kibana=kibana)
        self.types =["visualization", "dashboard", "search", "index-pattern", 
        "config", "timelion-sheet", "url", "query", "canvas-element", "canvas-workpad", "lens",
        "infrastructure-ui-source", "metrics-explorer-view", "inventory-view"]
        self.create_url = "api/saved_objects"
        self.import_url = "api/saved_objects/_import"
        self.all_url = "api/saved_objects/_find"
        self.get_url = "api/saved_objects"
        self.attribs = attribs
        self.references = references
        self.type = type.lower()

    def get(self, id, type=""):
        type = self.type if not type else type.lower()
        if not self.validate_type(type, types=self.types):
            return None
        url = self.url(self.get_url, type, id)
        return self.requester(url=url, method="get")

    def all(self, type=""):
        params = {
            "type": self.types if not type else type
        }
        url = self.url(self.all_url)
        return self.requester(url=url, method="get", params=params)

    def create(self, type="", attribs={}, references={}, body={}):
        type = (self.type if not type else type.lower())
        attribs = (self.attribs if not attribs else attribs)
        references = (self.references if not references else references)
        if not self.validate_type(type, types=self.types):
            return None
        url = self.url(self.create_url, type)
        data = {}
        if attribs:
            data["attributes"] = attribs
        if references:
            data["references"] = references
        data = data if not body else body
        response = self.requester(url=url, method="post", data=json.dumps(data))
        return response

    def loads(self, file):
        url = self.url(self.import_url)
        response = self.requester(url=url, method="post", files={"file": file})
        return response

    def dumps(self):
        pass

class Panel():
    def __init__(self, panel_name, width, height, pos_x, pos_y, id=None, visualization_id= "") -> None:
        self.id = self.__random_string() if not id else id
        self.panel_name = panel_name
        self.width = width
        self.height = height
        self.pos_x = pos_x
        self.pos_y = pos_y
        self.visualization_id=visualization_id

    def __random_string(self):
        return secrets.token_urlsafe(32)

    def create(self):
        return {
            "version": "7.8.0",
            "gridData": {
                "x": self.pos_x,
                "y": self.pos_y,
                "w": self.width,
                "h": self.height,
                "i": self.id
            },
            "panelIndex": self.id,
            "embeddableConfig": {},
            "panelRefName": self.panel_name
        }

    def get_reference(self):
        return {} if not self.visualization_id else {
            "name": self.panel_name,
            "type": "visualization",
            "id": self.visualization_id
        }

class Dashboard():
    def __init__(self, title, panels=[], references=[], query="") -> None:
        self.primitive_type="dashboard"
        self.title = title
        self.panels = panels
        self.references = references
        self.query=query

    def create(self):
        return {
            "attributes": {
                "title": self.title,
                "hits": 0,
                "description": "",
                "panelsJSON": json.dumps(self.panels),
                "optionsJSON": "{\"useMargins\":true,\"hidePanelTitles\":false}",
                "version": 1,
                "timeRestore": False,
                "kibanaSavedObjectMeta": {
                    "searchSourceJSON": self.__querier()
                }
            },
            "references": self.references
        }
        
    def __querier(self):
        data = {
            "query": {
                "query": self.query,
                "language": "kuery"
            },
            "filter": []
        }
        return json.dumps(data)
    
class Visualization(Utils):
    
    def __init__(self, index_pattern_id:str, title:str="", query:str="", mappings_dir_path:str=None, type:str="") -> None:
        CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))
        self.primitive_type="visualization"
        self.types=["area", "heatmap", "line", "metric", "pie", "table", "tagcloud"]
        self.type= type.lower()
        self.title = title
        self.index_pattern_id = index_pattern_id
        self.query = query
        self.mappings_dir_path = os.path.join(CURRENT_DIR, 'mappings') if not mappings_dir_path else mappings_dir_path
        self.data = self.load_visualizations()

    def load_visualizations(self):
        data = {}
        for type in self.types:
            file_path =  os.path.join(self.mappings_dir_path, "{}.json".format(type))
            data[type] = self.__read_json_file(file_path)
        return data

    def create(self, index_pattern_id:str=None, title:str="", body={}, query:str="") :
        title = self.title if not title else title
        visualization_state = self.__templater(title) if not body else self.__templater_json(title, body) 
        search_state = self.__querier(self.query if not query else query)
        index_pattern_id = self.index_pattern_id if not index_pattern_id else index_pattern_id
        return {
            "attributes": {
                "title": title,
                "visState": visualization_state,
                "uiStateJSON": "{}",
                "description": "",
                "version": 1,
                "kibanaSavedObjectMeta": {
                    "searchSourceJSON": search_state
                }
            },
            "references": [{
                "name": "kibanaSavedObjectMeta.searchSourceJSON.index",
                "type": "index-pattern",
                "id": index_pattern_id
            }]
        }

    def __read_json_file(self, file_path):
        file = open(file_path, 'r')
        read = file.read()
        file.close()
        return json.loads(read)

    def __templater_json(self, title, _json) -> str:
        data = _json
        data["title"] = title
        return json.dumps(data)

    def __templater(self, title) -> str:
        if self.validate_type(self.type, self.types):
            data = self.data[self.type]
            data["title"] = title
            return json.dumps(data)

    def __querier(self, query=None) -> str:
        data = {
            "query": {
                "query": query,
                "language": "kuery"
            },
            "indexRefName": "kibanaSavedObjectMeta.searchSourceJSON.index",
            "filter": []
        }
        return json.dumps(data)



# class TooLargePanel(Exception):
#     def __init__(self, *args: object) -> None:
#         super().__init__(*args)




import setuptools

with open("README.md", "r", encoding="utf-8") as description:
    long_description = description.read()

setuptools.setup(
    name="kibana-api",
    version="0.0.3",
    author="Mauricio Matias Conde",
    author_email="mcm.crw@gmail.com",
    description="This is an API mapping library for Kibana API to generate visualizations and dashboards automatically",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license='MIT',
    url="https://github.com/cr0wg4n/kibana-api",
    project_urls={
        "Bug Tracker": "https://github.com/cr0wg4n/kibana-api/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
    ],
    keywords=["elasticsearch", "kibana", "development", "api mapping"],
    packages=["kibana_api"],
    package_data={
        'kibana_api': [
            'mappings/*.json',
        ]
    },
    install_requires=[
        "requests",
        "multipledispatch"
    ],
    python_requires=">=3.0",
)


import requests
from urllib.parse import urljoin
from .objects import Space, Object

class Kibana:
    def __init__(self, base_url, username=None, password=None) -> None:
        self.base_url = base_url
        self.username = username
        self.password = password

    def ping(self):
        status_code = self.requester(method="get", url=self.base_url).status_code
        return True if status_code == 200 else False
        
    def url(self, *args) -> str:
        url = self.base_url
        for arg in args:
            url = urljoin(url, arg)+ "/"
        return  url[:-1] if url[-1:]=="/" else url

    def requester(self, **kwargs): 
        headers = {
            "Content-Type": "application/json",
            "kbn-xsrf": "True",
        } if not "files" in kwargs else {
            "kbn-xsrf": "True",
        }
        auth = (self.username, self.password) if (self.username and self.password) else None
        return requests.request(headers=headers, auth=auth, **kwargs)

    def space(self, **kwargs):
        return Space(kibana=self, **kwargs)

    def object(self, **kwargs): 
        return Object(kibana=self, **kwargs)
